# -*- coding: utf-8 -*-
"""
OnionShare | https://onionshare.org/

Copyright (C) 2014-2022 Micah Lee, et al. <micah@micahflee.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import time
import subprocess
import os
from datetime import datetime
from PySide6 import QtCore, QtWidgets, QtGui
from urllib.parse import unquote

from ... import strings
from ...widgets import Alert
from ...gui_common import GuiCommon


class HistoryItem(QtWidgets.QWidget):
    """
    The base history item
    """

    STATUS_STARTED = 0
    STATUS_FINISHED = 1
    STATUS_CANCELED = 2

    def __init__(self):
        super(HistoryItem, self).__init__()

    def update(self):
        pass

    def cancel(self):
        pass

    def get_finished_label_text(self, started):
        """
        When an item finishes, returns a string displaying the start/end datetime range.
        started is a datetime object.
        """
        return self._get_label_text(
            "gui_all_modes_transfer_finished",
            "gui_all_modes_transfer_finished_range",
            started,
        )

    def get_canceled_label_text(self, started):
        """
        When an item is canceled, returns a string displaying the start/end datetime range.
        started is a datetime object.
        """
        return self._get_label_text(
            "gui_all_modes_transfer_canceled",
            "gui_all_modes_transfer_canceled_range",
            started,
        )

    def _get_label_text(self, string_name, string_range_name, started):
        """
        Return a string that contains a date, or date range.
        """
        ended = datetime.now()
        if (
            started.year == ended.year
            and started.month == ended.month
            and started.day == ended.day
        ):
            if started.hour == ended.hour and started.minute == ended.minute:
                text = strings._(string_name).format(started.strftime("%b %d, %I:%M%p"))
            else:
                text = strings._(string_range_name).format(
                    started.strftime("%b %d, %I:%M%p"), ended.strftime("%I:%M%p")
                )
        else:
            text = strings._(string_range_name).format(
                started.strftime("%b %d, %I:%M%p"), ended.strftime("%b %d, %I:%M%p")
            )
        return text


class ShareHistoryItem(HistoryItem):
    """
    Download history item, for share mode
    """

    def __init__(self, common, id, total_bytes):
        super(ShareHistoryItem, self).__init__()
        self.common = common

        self.id = id
        self.total_bytes = total_bytes
        self.downloaded_bytes = 0
        self.started = time.time()
        self.started_dt = datetime.fromtimestamp(self.started)
        self.status = HistoryItem.STATUS_STARTED

        # Label
        self.label = QtWidgets.QLabel(
            strings._("gui_all_modes_transfer_started").format(
                self.started_dt.strftime("%b %d, %I:%M%p")
            )
        )

        # Progress bar
        self.progress_bar = QtWidgets.QProgressBar()
        self.progress_bar.setTextVisible(True)
        self.progress_bar.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        self.progress_bar.setAlignment(QtCore.Qt.AlignHCenter)
        self.progress_bar.setMinimum(0)
        self.progress_bar.setMaximum(total_bytes / 1024)
        self.progress_bar.setValue(0)
        self.progress_bar.setStyleSheet(
            self.common.gui.css["downloads_uploads_progress_bar"]
        )
        self.progress_bar.total_bytes = total_bytes

        # Layout
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.label)
        layout.addWidget(self.progress_bar)
        self.setLayout(layout)

        # Start at 0
        self.update(0)

    def update(self, downloaded_bytes):
        self.downloaded_bytes = downloaded_bytes

        self.progress_bar.setValue(downloaded_bytes / 1024)
        if (downloaded_bytes / 1024) == (self.progress_bar.total_bytes / 1024):
            pb_fmt = strings._("gui_all_modes_progress_complete").format(
                self.common.format_seconds(time.time() - self.started)
            )

            # Change the label
            self.label.setText(self.get_finished_label_text(self.started_dt))
            self.label.setStyleSheet(self.common.gui.css["history_default_label"])
            self.status = HistoryItem.STATUS_FINISHED

        else:
            elapsed = time.time() - self.started
            if elapsed < 10:
                # Wait a couple of seconds for the download rate to stabilize.
                # This prevents a "Windows copy dialog"-esque experience at
                # the beginning of the download.
                pb_fmt = strings._("gui_all_modes_progress_starting").format(
                    self.common.human_readable_filesize(downloaded_bytes)
                )
            else:
                pb_fmt = strings._("gui_all_modes_progress_eta").format(
                    self.common.human_readable_filesize(downloaded_bytes),
                    self.estimated_time_remaining,
                )

        self.progress_bar.setFormat(pb_fmt)

    def cancel(self):
        self.progress_bar.setFormat(strings._("gui_canceled"))
        self.status = HistoryItem.STATUS_CANCELED

    @property
    def estimated_time_remaining(self):
        return self.common.estimated_time_remaining(
            self.downloaded_bytes, self.total_bytes, self.started
        )


class ReceiveHistoryItemFile(QtWidgets.QWidget):
    def __init__(self, common, filename):
        super(ReceiveHistoryItemFile, self).__init__()
        self.common = common

        self.common.log("ReceiveHistoryItemFile", "__init__", f"filename: {filename}")

        self.filename = filename
        self.dir = None
        self.started = datetime.now()

        # Filename label
        self.filename_label = QtWidgets.QLabel(self.filename)
        self.filename_label_width = self.filename_label.width()

        # File size label
        self.filesize_label = QtWidgets.QLabel()
        self.filesize_label.setStyleSheet(self.common.gui.css["receive_file_size"])
        self.filesize_label.hide()

        # Folder button
        image = QtGui.QImage(GuiCommon.get_resource_path("images/open_folder.svg"))
        scaled_image = image.scaledToHeight(15, QtCore.Qt.SmoothTransformation)
        folder_pixmap = QtGui.QPixmap.fromImage(scaled_image)
        folder_icon = QtGui.QIcon(folder_pixmap)
        self.folder_button = QtWidgets.QPushButton()
        self.folder_button.clicked.connect(self.open_folder)
        self.folder_button.setIcon(folder_icon)
        self.folder_button.setIconSize(folder_pixmap.rect().size())
        self.folder_button.setFlat(True)
        self.folder_button.hide()

        # Layouts
        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(self.filename_label)
        layout.addWidget(self.filesize_label)
        layout.addStretch()
        layout.addWidget(self.folder_button)
        self.setLayout(layout)

    def update(self, uploaded_bytes, complete):
        self.filesize_label.setText(self.common.human_readable_filesize(uploaded_bytes))
        self.filesize_label.show()

        if complete:
            self.folder_button.show()

    def rename(self, new_filename):
        self.filename = new_filename
        self.filename_label.setText(self.filename)

    def set_dir(self, dir):
        self.dir = dir

    def open_folder(self):
        """
        Open the downloads folder, with the file selected, in a cross-platform manner
        """
        self.common.log("ReceiveHistoryItemFile", "open_folder")

        if not self.dir:
            self.common.log(
                "ReceiveHistoryItemFile",
                "open_folder",
                "dir has not been set yet, can't open folder",
            )
            return

        abs_filename = os.path.join(self.dir, self.filename)

        # Linux
        if self.common.platform == "Linux" or self.common.platform == "BSD":
            try:
                # If nautilus is available, open it
                subprocess.Popen(["xdg-open", self.dir])
            except Exception:
                Alert(
                    self.common,
                    strings._("gui_open_folder_error").format(abs_filename),
                )

        # macOS
        elif self.common.platform == "Darwin":
            subprocess.call(["open", "-R", abs_filename])

        # Windows
        elif self.common.platform == "Windows":
            subprocess.Popen(["explorer", f"/select,{abs_filename}"])


class ReceiveHistoryItemMessage(QtWidgets.QWidget):
    def __init__(
        self,
        common,
    ):
        super(ReceiveHistoryItemMessage, self).__init__()
        self.common = common
        self.filename = None

        # Read message button
        message_pixmap = QtGui.QPixmap.fromImage(
            QtGui.QImage(GuiCommon.get_resource_path("images/open_message.png"))
        )
        message_icon = QtGui.QIcon(message_pixmap)
        self.message_button = QtWidgets.QPushButton(
            strings._("history_receive_read_message_button")
        )
        self.message_button.setStyleSheet(self.common.gui.css["receive_message_button"])
        self.message_button.clicked.connect(self.open_message)
        self.message_button.setIcon(message_icon)
        self.message_button.setIconSize(message_pixmap.rect().size())

        # Layouts
        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(self.message_button)
        layout.addStretch()
        self.setLayout(layout)

        self.hide()

    def set_filename(self, new_filename):
        self.filename = new_filename
        self.show()

    def open_message(self):
        """
        Open the message in the operating system's default text editor
        """
        self.common.log("ReceiveHistoryItemMessage", "open_message", self.filename)

        # Linux
        if self.common.platform == "Linux" or self.common.platform == "BSD":
            # If nautilus is available, open it
            subprocess.Popen(["xdg-open", self.filename])

        # macOS
        elif self.common.platform == "Darwin":
            subprocess.call(["open", self.filename])

        # Windows
        elif self.common.platform == "Windows":
            subprocess.Popen(["notepad", self.filename])


class ReceiveHistoryItem(HistoryItem):
    def __init__(self, common, id, content_length):
        super(ReceiveHistoryItem, self).__init__()
        self.common = common
        self.id = id
        self.content_length = content_length
        self.started = datetime.now()
        self.status = HistoryItem.STATUS_STARTED

        self.common.log(
            "ReceiveHistoryItem",
            "__init__",
            f"id={self.id} content_length={self.content_length}",
        )

        # Label
        self.label = QtWidgets.QLabel(
            strings._("gui_all_modes_transfer_started").format(
                self.started.strftime("%b %d, %I:%M%p")
            )
        )

        # Progress bar
        self.progress_bar = QtWidgets.QProgressBar()
        self.progress_bar.setTextVisible(True)
        self.progress_bar.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        self.progress_bar.setAlignment(QtCore.Qt.AlignHCenter)
        self.progress_bar.setMinimum(0)
        self.progress_bar.setValue(0)
        self.progress_bar.setStyleSheet(
            self.common.gui.css["downloads_uploads_progress_bar"]
        )

        # The message widget, if a message was included
        self.message = ReceiveHistoryItemMessage(self.common)

        # This layout contains file widgets
        self.files_layout = QtWidgets.QVBoxLayout()
        self.files_layout.setContentsMargins(0, 0, 0, 0)
        files_widget = QtWidgets.QWidget()
        files_widget.setStyleSheet(self.common.gui.css["receive_file"])
        files_widget.setLayout(self.files_layout)

        # Layout
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.label)
        layout.addWidget(self.message)
        layout.addWidget(self.progress_bar)
        layout.addWidget(files_widget)
        layout.addStretch()
        self.setLayout(layout)

        # We're also making a dictionary of file widgets, to make them easier to access
        self.files = {}

    def includes_message(self, message_filename):
        self.message.set_filename(message_filename)

    def update(self, data):
        """
        Using the progress from Web, update the progress bar and file size labels
        for each file
        """
        if data["action"] == "progress":
            total_uploaded_bytes = 0
            for filename in data["progress"]:
                total_uploaded_bytes += data["progress"][filename]["uploaded_bytes"]

            # Update the progress bar
            self.progress_bar.setMaximum(self.content_length / 1024)
            self.progress_bar.setValue(total_uploaded_bytes / 1024)

            elapsed = datetime.now() - self.started
            if elapsed.seconds < 10:
                pb_fmt = strings._("gui_all_modes_progress_starting").format(
                    self.common.human_readable_filesize(total_uploaded_bytes)
                )
            else:
                estimated_time_remaining = self.common.estimated_time_remaining(
                    total_uploaded_bytes, self.content_length, self.started.timestamp()
                )
                pb_fmt = strings._("gui_all_modes_progress_eta").format(
                    self.common.human_readable_filesize(total_uploaded_bytes),
                    estimated_time_remaining,
                )

            self.progress_bar.setFormat(pb_fmt)

            # Using list(progress) to avoid "RuntimeError: dictionary changed size during iteration"
            for filename in list(data["progress"]):
                # Add a new file if needed
                if filename not in self.files:
                    self.files[filename] = ReceiveHistoryItemFile(self.common, filename)
                    self.files_layout.addWidget(self.files[filename])

                # Update the file
                self.files[filename].update(
                    data["progress"][filename]["uploaded_bytes"],
                    data["progress"][filename]["complete"],
                )

        elif data["action"] == "rename":
            self.files[data["old_filename"]].rename(data["new_filename"])
            self.files[data["new_filename"]] = self.files.pop(data["old_filename"])

        elif data["action"] == "set_dir":
            self.files[data["filename"]].set_dir(data["dir"])

        elif data["action"] == "finished":
            # Change the status
            self.status = HistoryItem.STATUS_FINISHED

            # Hide the progress bar
            self.progress_bar.hide()

            # Change the label
            self.label.setText(self.get_finished_label_text(self.started))
            self.label.setStyleSheet(self.common.gui.css["history_default_label"])

        elif data["action"] == "canceled":
            # Change the status
            self.status = HistoryItem.STATUS_CANCELED

            # Hide the progress bar
            self.progress_bar.hide()

            # Change the label
            self.label.setText(self.get_canceled_label_text(self.started))


class IndividualFileHistoryItem(HistoryItem):
    """
    Individual file history item, for share mode viewing of individual files
    """

    def __init__(self, common, data, path):
        super(IndividualFileHistoryItem, self).__init__()
        self.status = HistoryItem.STATUS_STARTED
        self.common = common

        self.id = id
        # Decode and sanitize the path to remove newlines
        decoded_path = unquote(path)
        self.path = decoded_path.replace("\r", "").replace("\n", "")
        self.total_bytes = 0
        self.downloaded_bytes = 0
        self.started = time.time()
        self.started_dt = datetime.fromtimestamp(self.started)
        self.status = HistoryItem.STATUS_STARTED

        self.directory_listing = "directory_listing" in data

        # Labels
        self.timestamp_label = QtWidgets.QLabel(
            self.started_dt.strftime("%b %d, %I:%M%p")
        )
        self.timestamp_label.setStyleSheet(
            self.common.gui.css["history_individual_file_timestamp_label"]
        )
        self.path_label = QtWidgets.QLabel(self.path)
        self.path_label.setTextFormat(QtCore.Qt.PlainText)
        self.path_label.setStyleSheet(self.common.gui.css["history_default_label"])
        self.status_code_label = QtWidgets.QLabel()

        # Progress bar
        self.progress_bar = QtWidgets.QProgressBar()
        self.progress_bar.setTextVisible(True)
        self.progress_bar.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        self.progress_bar.setAlignment(QtCore.Qt.AlignHCenter)
        self.progress_bar.setValue(0)
        self.progress_bar.setStyleSheet(
            self.common.gui.css["downloads_uploads_progress_bar"]
        )

        # Text layout
        labels_layout = QtWidgets.QHBoxLayout()
        labels_layout.addWidget(self.timestamp_label)
        labels_layout.addWidget(self.path_label)
        labels_layout.addWidget(self.status_code_label)
        labels_layout.addStretch()

        # Layout
        layout = QtWidgets.QVBoxLayout()
        layout.addLayout(labels_layout)
        layout.addWidget(self.progress_bar)
        self.setLayout(layout)

        # Is a status code already sent?
        if "status_code" in data:
            self.status_code_label.setText(str(data["status_code"]))
            if data["status_code"] >= 200 and data["status_code"] < 300:
                self.status_code_label.setStyleSheet(
                    self.common.gui.css["history_individual_file_status_code_label_2xx"]
                )
            if data["status_code"] >= 400 and data["status_code"] < 500:
                self.status_code_label.setStyleSheet(
                    self.common.gui.css["history_individual_file_status_code_label_4xx"]
                )
            self.status = HistoryItem.STATUS_FINISHED
            self.progress_bar.hide()
            return

        else:
            self.total_bytes = data["filesize"]
            self.progress_bar.setMinimum(0)
            self.progress_bar.setMaximum(data["filesize"] / 1024)
            self.progress_bar.total_bytes = data["filesize"]

        # Start at 0
        self.update(0)

    def update(self, downloaded_bytes):
        self.downloaded_bytes = downloaded_bytes

        self.progress_bar.setValue(downloaded_bytes / 1024)
        if (downloaded_bytes / 1024) == (self.progress_bar.total_bytes / 1024):
            self.status_code_label.setText("200")
            self.status_code_label.setStyleSheet(
                self.common.gui.css["history_individual_file_status_code_label_2xx"]
            )
            self.progress_bar.hide()
            self.status = HistoryItem.STATUS_FINISHED

        else:
            elapsed = time.time() - self.started
            if elapsed < 10:
                # Wait a couple of seconds for the download rate to stabilize.
                # This prevents a "Windows copy dialog"-esque experience at
                # the beginning of the download.
                pb_fmt = strings._("gui_all_modes_progress_starting").format(
                    self.common.human_readable_filesize(downloaded_bytes)
                )
            else:
                pb_fmt = strings._("gui_all_modes_progress_eta").format(
                    self.common.human_readable_filesize(downloaded_bytes),
                    self.estimated_time_remaining,
                )

            self.progress_bar.setFormat(pb_fmt)

    def cancel(self):
        self.progress_bar.setFormat(strings._("gui_canceled"))
        self.status = HistoryItem.STATUS_CANCELED

    @property
    def estimated_time_remaining(self):
        return self.common.estimated_time_remaining(
            self.downloaded_bytes, self.total_bytes, self.started
        )


class HistoryItemList(QtWidgets.QScrollArea):
    """
    List of items
    """

    def __init__(self, common):
        super(HistoryItemList, self).__init__()
        self.common = common

        self.items = {}

        # The layout that holds all of the items
        self.items_layout = QtWidgets.QVBoxLayout()
        self.items_layout.setContentsMargins(0, 0, 0, 0)
        self.items_layout.setSizeConstraint(QtWidgets.QLayout.SetMinAndMaxSize)

        # Wrapper layout that also contains a stretch
        wrapper_layout = QtWidgets.QVBoxLayout()
        wrapper_layout.setSizeConstraint(QtWidgets.QLayout.SetMinAndMaxSize)
        wrapper_layout.addLayout(self.items_layout)
        wrapper_layout.addStretch()

        # The internal widget of the scroll area
        widget = QtWidgets.QWidget()
        widget.setLayout(wrapper_layout)
        self.setWidget(widget)
        self.setWidgetResizable(True)

        # Other scroll area settings
        self.setBackgroundRole(QtGui.QPalette.Light)
        self.verticalScrollBar().rangeChanged.connect(self.resizeScroll)

    def resizeScroll(self, minimum, maximum):
        """
        Scroll to the bottom of the window when the range changes.
        """
        self.verticalScrollBar().setValue(maximum)

    def add(self, id, item):
        """
        Add a new item. Override this method.
        """
        self.items[id] = item
        self.items_layout.addWidget(item)

    def update(self, id, data):
        """
        Update an item.  Override this method.
        """
        if id in self.items:
            self.items[id].update(data)

    def cancel(self, id):
        """
        Cancel an item.  Override this method.
        """
        if id in self.items:
            self.items[id].cancel()

    def includes_message(self, id, message_filename):
        """
        Show message button for receive mode
        """
        if id in self.items:
            self.items[id].includes_message(message_filename)

    def reset(self):
        """
        Reset all items, emptying the list.  Override this method.
        """
        for key, item in self.items.copy().items():
            self.items_layout.removeWidget(item)
            item.close()
            del self.items[key]


class History(QtWidgets.QWidget):
    """
    A history of what's happened so far in this mode. This contains an internal
    object full of a scrollable list of items.
    """

    def __init__(self, common, empty_image, empty_text, header_text, mode=""):
        super(History, self).__init__()
        self.common = common
        self.mode = mode

        self.setMinimumWidth(350)

        # In progress and completed counters
        self.in_progress_count = 0
        self.completed_count = 0
        self.requests_count = 0

        # In progress, completed, and requests labels
        self.in_progress_label = QtWidgets.QLabel()
        self.in_progress_label.setStyleSheet(self.common.gui.css["mode_info_label"])
        self.completed_label = QtWidgets.QLabel()
        self.completed_label.setStyleSheet(self.common.gui.css["mode_info_label"])
        self.requests_label = QtWidgets.QLabel()
        self.requests_label.setStyleSheet(self.common.gui.css["mode_info_label"])

        # Header
        self.header_label = QtWidgets.QLabel(header_text)
        self.header_label.setStyleSheet(self.common.gui.css["downloads_uploads_label"])
        self.clear_button = QtWidgets.QPushButton(
            strings._("gui_all_modes_clear_history")
        )
        self.clear_button.setStyleSheet(self.common.gui.css["downloads_uploads_clear"])
        self.clear_button.setFlat(True)
        self.clear_button.clicked.connect(self.reset)
        header_layout = QtWidgets.QHBoxLayout()
        header_layout.addWidget(self.header_label)
        header_layout.addStretch()
        header_layout.addWidget(self.in_progress_label)
        header_layout.addWidget(self.completed_label)
        header_layout.addWidget(self.requests_label)
        header_layout.addWidget(self.clear_button)

        # When there are no items
        self.empty_image = QtWidgets.QLabel()
        self.empty_image.setAlignment(QtCore.Qt.AlignCenter)
        self.empty_image.setPixmap(empty_image)
        self.empty_text = QtWidgets.QLabel(empty_text)
        self.empty_text.setAlignment(QtCore.Qt.AlignCenter)
        self.empty_text.setStyleSheet(
            self.common.gui.css["downloads_uploads_empty_text"]
        )
        empty_layout = QtWidgets.QVBoxLayout()
        empty_layout.addStretch()
        empty_layout.addWidget(self.empty_image)
        empty_layout.addWidget(self.empty_text)
        empty_layout.addStretch()
        self.empty = QtWidgets.QWidget()
        self.empty.setStyleSheet(self.common.gui.css["downloads_uploads_empty"])
        self.empty.setLayout(empty_layout)

        # When there are items
        self.item_list = HistoryItemList(self.common)
        self.not_empty_layout = QtWidgets.QVBoxLayout()
        self.not_empty_layout.addLayout(header_layout)
        self.not_empty_layout.addWidget(self.item_list)
        self.not_empty = QtWidgets.QWidget()
        self.not_empty.setStyleSheet(self.common.gui.css["downloads_uploads_not_empty"])
        self.not_empty.setLayout(self.not_empty_layout)

        # Layout
        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.empty)
        layout.addWidget(self.not_empty)
        self.setLayout(layout)

        # Reset once at the beginning
        self.reset()

    def add(self, id, item):
        """
        Add a new item.
        """
        self.common.log("History", "add", f"id: {id}")

        # Hide empty, show not empty
        self.empty.hide()
        self.not_empty.show()

        # Add it to the list
        self.item_list.add(id, item)

    def update(self, id, data):
        """
        Update an item.
        """
        self.item_list.update(id, data)

    def cancel(self, id):
        """
        Cancel an item.
        """
        self.item_list.cancel(id)

    def includes_message(self, id, message_filename):
        """
        Show the message button
        """
        self.item_list.includes_message(id, message_filename)

    def reset(self):
        """
        Reset all items.
        """
        self.item_list.reset()
        if len(self.item_list.items) == 0:
            # Hide not empty, show empty
            self.not_empty.hide()
            self.empty.show()
            # Reset in-progress counter
            self.in_progress_count = 0
            self.update_in_progress()

        # Reset completed counter
        self.completed_count = 0
        self.update_completed()

        # Reset web requests counter
        self.requests_count = 0
        self.update_requests()

    def update_completed(self):
        """
        Update the 'completed' widget.
        """
        if self.completed_count == 0:
            image = GuiCommon.get_resource_path(f"images/{self.common.gui.color_mode}_history_completed_none.svg")
        else:
            image = GuiCommon.get_resource_path("images/history_completed.svg")
        self.completed_label.setText(f'<img src="{image}" height="10" /> {self.completed_count}')
        self.completed_label.setToolTip(
            strings._("history_completed_tooltip").format(self.completed_count)
        )

    def update_in_progress(self):
        """
        Update the 'in progress' widget.
        """
        if self.in_progress_count == 0:
            image = GuiCommon.get_resource_path(f"images/{self.common.gui.color_mode}_history_in_progress_none.svg")
        else:
            image = GuiCommon.get_resource_path("images/history_in_progress.svg")

        self.in_progress_label.setText(
            f'<img src="{image}" height="10" /> {self.in_progress_count}'
        )
        self.in_progress_label.setToolTip(
            strings._("history_in_progress_tooltip").format(self.in_progress_count)
        )

    def update_requests(self):
        """
        Update the 'web requests' widget.
        """
        if self.requests_count == 0:
            image = GuiCommon.get_resource_path(f"images/{self.common.gui.color_mode}_history_requests_none.svg")
        else:
            image = GuiCommon.get_resource_path("images/history_requests.svg")

        self.requests_label.setText(f'<img src="{image}" height="10" /> {self.requests_count}')
        self.requests_label.setToolTip(
            strings._("history_requests_tooltip").format(self.requests_count)
        )


class ToggleHistory(QtWidgets.QPushButton):
    """
    Widget for toggling showing or hiding the history, as well as keeping track
    of the indicator counter if it's hidden
    """

    def __init__(self, common, current_mode, history_widget, icon, selected_icon):
        super(ToggleHistory, self).__init__()
        self.common = common
        self.current_mode = current_mode
        self.history_widget = history_widget
        self.icon = icon
        self.selected_icon = selected_icon

        # Toggle button
        self.setDefault(False)
        self.setFixedWidth(35)
        self.setFixedHeight(30)
        self.setFlat(True)
        self.setIcon(icon)
        self.clicked.connect(self.toggle_clicked)

        # Keep track of indicator
        self.indicator_count = 0
        self.indicator_label = QtWidgets.QLabel(parent=self)
        self.indicator_label.setStyleSheet(
            self.common.gui.css["download_uploads_indicator"]
        )
        self.update_indicator()

    def update_indicator(self, increment=False):
        """
        Update the display of the indicator count. If increment is True, then
        only increment the counter if History is hidden.
        """
        if increment and not self.history_widget.isVisible():
            self.indicator_count += 1

        self.indicator_label.setText(str(self.indicator_count))

        if self.indicator_count == 0:
            self.indicator_label.hide()
        else:
            size = self.indicator_label.sizeHint()
            self.indicator_label.setGeometry(
                35 - size.width(), 0, size.width(), size.height()
            )
            self.indicator_label.show()

    def toggle_clicked(self):
        """
        Toggle showing and hiding the history widget
        """
        self.common.log("ToggleHistory", "toggle_clicked")

        if self.history_widget.isVisible():
            self.history_widget.hide()
            self.setIcon(self.icon)
            self.setFlat(True)
        else:
            self.history_widget.show()
            self.setIcon(self.selected_icon)
            self.setFlat(False)

        # Reset the indicator count
        self.indicator_count = 0
        self.update_indicator()



from PySide6 import QtCore, QtWidgets, QtGui

from onionshare_cli.mode_settings import ModeSettings

from . import strings
from .tab import Tab
from .threads import EventHandlerThread
from .gui_common import GuiCommon
from .tor_settings_tab import TorSettingsTab
from .settings_tab import SettingsTab
from .connection_tab import AutoConnectTab


class SettingsParentTab(QtWidgets.QTabWidget):
    """
    The settings tab widget containing the tor settings
    and app settings subtabs
    """

    bring_to_front = QtCore.Signal()
    close_this_tab = QtCore.Signal()

    def __init__(
        self, common, tab_id, parent=None, active_tab="general", from_autoconnect=False
    ):
        super(SettingsParentTab, self).__init__()
        self.parent = parent
        self.common = common
        self.common.log("SettingsParentTab", "__init__")

        # Keep track of tabs in a dictionary that maps tab_id to tab.
        # Each tab has a unique, auto-incremented id (tab_id). This is different than the
        # tab's index, which changes as tabs are re-arranged.
        self.tabs = {
            "general": 0,
            "tor": 1,
        }
        self.tab_id = tab_id
        self.current_tab_id = self.tabs[active_tab]

        # Use a custom tab bar
        tab_bar = TabBar(self.common)
        self.setTabBar(tab_bar)
        settings_tab = SettingsTab(self.common, self.tabs["general"], parent=self)
        self.tor_settings_tab = TorSettingsTab(
            self.common,
            self.tabs["tor"],
            self.parent.are_tabs_active(),
            self.parent.status_bar,
            parent=self,
            from_autoconnect=from_autoconnect,
        )
        self.addTab(settings_tab, strings._("gui_general_settings_window_title"))
        self.addTab(self.tor_settings_tab, strings._("gui_tor_settings_window_title"))

        # Set up the tab widget
        self.setMovable(False)
        self.setTabsClosable(False)
        self.setUsesScrollButtons(False)
        self.setCurrentIndex(self.current_tab_id)


class TabBar(QtWidgets.QTabBar):
    """
    A custom tab bar
    """

    move_new_tab_button = QtCore.Signal()

    def __init__(self, common):
        super(TabBar, self).__init__()
        self.setStyleSheet(common.gui.css["settings_subtab_bar"])



# -*- coding: utf-8 -*-
"""
OnionShare | https://onionshare.org/

Copyright (C) 2014-2022 Micah Lee, et al. <micah@micahflee.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import os

from PySide6 import QtCore, QtWidgets, QtGui

from onionshare_cli.common import Common
from onionshare_cli.web import Web

from .. import Mode
from ..file_selection import FileSelection
from ..history import History, ToggleHistory
from .... import strings
from ....widgets import MinimumSizeWidget
from ....gui_common import GuiCommon


class WebsiteMode(Mode):
    """
    Parts of the main window UI for sharing files.
    """

    success = QtCore.Signal()
    error = QtCore.Signal(str)

    def init(self):
        """
        Custom initialization for ReceiveMode.
        """
        # Create the Web object
        self.web = Web(self.common, True, self.settings, "website")

        # Settings
        # Disable CSP option
        self.disable_csp_checkbox = QtWidgets.QCheckBox()
        self.disable_csp_checkbox.clicked.connect(self.disable_csp_checkbox_clicked)
        self.disable_csp_checkbox.setText(
            strings._("mode_settings_website_disable_csp_checkbox")
        )
        if self.settings.get("website", "disable_csp"):
            self.disable_csp_checkbox.setCheckState(QtCore.Qt.Checked)
        else:
            self.disable_csp_checkbox.setCheckState(QtCore.Qt.Unchecked)

        self.mode_settings_widget.mode_specific_layout.addWidget(
            self.disable_csp_checkbox
        )

        # Custom CSP option
        self.custom_csp_checkbox = QtWidgets.QCheckBox()
        self.custom_csp_checkbox.clicked.connect(self.custom_csp_checkbox_clicked)
        self.custom_csp_checkbox.setText(strings._("mode_settings_website_custom_csp_checkbox"))
        if self.settings.get("website", "custom_csp") and not self.settings.get("website", "disable_csp"):
            self.custom_csp_checkbox.setCheckState(QtCore.Qt.Checked)
        else:
            self.custom_csp_checkbox.setCheckState(QtCore.Qt.Unchecked)
        self.custom_csp = QtWidgets.QLineEdit()
        self.custom_csp.setPlaceholderText(
            "default-src 'self'; frame-ancestors 'none'; form-action 'self'; base-uri 'self'; img-src 'self' data:;"
        )
        self.custom_csp.editingFinished.connect(self.custom_csp_editing_finished)

        custom_csp_layout = QtWidgets.QHBoxLayout()
        custom_csp_layout.setContentsMargins(0, 0, 0, 0)
        custom_csp_layout.addWidget(self.custom_csp_checkbox)
        custom_csp_layout.addWidget(self.custom_csp)
        self.mode_settings_widget.mode_specific_layout.addLayout(custom_csp_layout)

        # File selection
        self.file_selection = FileSelection(
            self.common,
            "images/{}_mode_website.png".format(self.common.gui.color_mode),
            strings._("gui_new_tab_website_button"),
            self,
        )
        if self.filenames:
            for filename in self.filenames:
                self.file_selection.file_list.add_file(filename)

        # Set title placeholder
        self.mode_settings_widget.title_lineedit.setPlaceholderText(
            strings._("gui_tab_name_website")
        )

        # Server status
        self.server_status.set_mode("website", self.file_selection)
        self.server_status.server_started.connect(self.file_selection.server_started)
        self.server_status.server_stopped.connect(self.file_selection.server_stopped)
        self.server_status.server_stopped.connect(self.update_primary_action)
        self.server_status.server_canceled.connect(self.file_selection.server_stopped)
        self.server_status.server_canceled.connect(self.update_primary_action)
        self.file_selection.file_list.files_updated.connect(self.server_status.update)
        self.file_selection.file_list.files_updated.connect(self.update_primary_action)
        # Tell server_status about web, then update
        self.server_status.web = self.web
        self.server_status.update()

        # Filesize warning
        self.filesize_warning = QtWidgets.QLabel()
        self.filesize_warning.setWordWrap(True)
        self.filesize_warning.setStyleSheet(
            self.common.gui.css["share_filesize_warning"]
        )
        self.filesize_warning.hide()

        # Download history
        self.history = History(
            self.common,
            QtGui.QPixmap.fromImage(
                QtGui.QImage(
                    GuiCommon.get_resource_path("images/share_icon_transparent.png")
                )
            ),
            strings._("gui_website_mode_no_files"),
            strings._("gui_all_modes_history"),
            "website",
        )
        self.history.in_progress_label.hide()
        self.history.completed_label.hide()
        self.history.hide()

        # Info label
        self.info_label = QtWidgets.QLabel()
        self.info_label.hide()

        # Delete all files button
        self.remove_all_button = QtWidgets.QPushButton(
            strings._("gui_file_selection_remove_all")
        )
        self.remove_all_button.setFlat(True)
        self.remove_all_button.setStyleSheet(
            self.common.gui.css["share_delete_all_files_button"]
        )
        self.remove_all_button.clicked.connect(self.delete_all)
        self.remove_all_button.hide()

        # Toggle history
        self.toggle_history = ToggleHistory(
            self.common,
            self,
            self.history,
            QtGui.QIcon(GuiCommon.get_resource_path(f"images/{self.common.gui.color_mode}_history_icon_toggle.svg")),
            QtGui.QIcon(
                GuiCommon.get_resource_path(f"images/{self.common.gui.color_mode}_history_icon_toggle_selected.svg")
            ),
        )

        # Top bar
        top_bar_layout = QtWidgets.QHBoxLayout()
        top_bar_layout.addWidget(self.info_label)
        top_bar_layout.addStretch()
        top_bar_layout.addWidget(self.remove_all_button)
        top_bar_layout.addWidget(self.toggle_history)

        # Primary action layout
        self.primary_action_layout.addWidget(self.filesize_warning)
        self.primary_action.hide()
        self.update_primary_action()

        # Main layout
        self.main_layout = QtWidgets.QVBoxLayout()
        self.main_layout.addLayout(top_bar_layout)
        self.main_layout.addLayout(self.file_selection)
        self.main_layout.addWidget(self.primary_action, stretch=1)
        self.main_layout.addWidget(self.server_status)
        self.main_layout.addWidget(MinimumSizeWidget(700, 0))

        # Column layout
        self.column_layout = QtWidgets.QHBoxLayout()
        self.column_layout.addLayout(self.main_layout)
        self.column_layout.addWidget(self.history, stretch=1)

        # Content layout
        self.content_layout.addLayout(self.column_layout)

        # Always start with focus on file selection
        self.file_selection.setFocus()

    def get_type(self):
        """
        Returns the type of mode as a string (e.g. "share", "receive", etc.)
        """
        return "website"

    def disable_csp_checkbox_clicked(self):
        """
        Save disable CSP setting to the tab settings. Uncheck 'custom CSP'
        setting if disabling CSP altogether.
        """
        self.settings.set(
            "website", "disable_csp", self.disable_csp_checkbox.isChecked()
        )
        if self.disable_csp_checkbox.isChecked():
            self.custom_csp_checkbox.setCheckState(QtCore.Qt.Unchecked)
            self.custom_csp_checkbox.setEnabled(False)
        else:
            self.custom_csp_checkbox.setEnabled(True)

    def custom_csp_checkbox_clicked(self):
        """
        Uncheck 'disable CSP' setting if custom CSP is used.
        """
        if self.custom_csp_checkbox.isChecked():
            self.disable_csp_checkbox.setCheckState(QtCore.Qt.Unchecked)
            self.disable_csp_checkbox.setEnabled(False)
            self.settings.set(
                "website", "custom_csp", self.custom_csp
            )
        else:
            self.disable_csp_checkbox.setEnabled(True)
            self.custom_csp.setText("")
            self.settings.set(
                "website", "custom_csp", None
            )

    def custom_csp_editing_finished(self):
        if self.custom_csp.text().strip() == "":
            self.custom_csp.setText("")
            self.settings.set("website", "custom_csp", None)
        else:
            custom_csp = self.custom_csp.text()
            self.settings.set("website", "custom_csp", custom_csp)

    def get_stop_server_autostop_timer_text(self):
        """
        Return the string to put on the stop server button, if there's an auto-stop timer
        """
        return strings._("gui_share_stop_server_autostop_timer")

    def autostop_timer_finished_should_stop_server(self):
        """
        The auto-stop timer expired, should we stop the server? Returns a bool
        """

        self.server_status.stop_server()
        self.server_status_label.setText(strings._("close_on_autostop_timer"))
        return True

    def start_server_custom(self):
        """
        Starting the server.
        """
        # Reset web counters
        self.web.website_mode.visit_count = 0

        # Hide and reset the downloads if we have previously shared
        self.reset_info_counters()

        self.remove_all_button.hide()

    def start_server_step2_custom(self):
        """
        Step 2 in starting the server. Zipping up files.
        """
        self.filenames = []
        for index in range(self.file_selection.file_list.count()):
            self.filenames.append(self.file_selection.file_list.item(index).filename)

        # Continue
        self.starting_server_step3.emit()
        self.start_server_finished.emit()

    def start_server_step3_custom(self):
        """
        Step 3 in starting the server. Display large filesize
        warning, if applicable.
        """
        self.web.website_mode.set_file_info(self.filenames)
        self.success.emit()

    def start_server_error_custom(self):
        """
        Start server error.
        """
        if self._zip_progress_bar is not None:
            self.status_bar.removeWidget(self._zip_progress_bar)
            self._zip_progress_bar = None

    def stop_server_custom(self):
        """
        Stop server.
        """

        self.filesize_warning.hide()
        self.history.completed_count = 0
        self.file_selection.file_list.adjustSize()

        self.remove_all_button.show()

    def cancel_server_custom(self):
        """
        Log that the server has been cancelled
        """
        self.common.log("WebsiteMode", "cancel_server")

    def handle_tor_broke_custom(self):
        """
        Connection to Tor broke.
        """
        self.primary_action.hide()

    def update_primary_action(self):
        self.common.log("WebsiteMode", "update_primary_action")

        # Show or hide primary action layout
        file_count = self.file_selection.file_list.count()
        if file_count > 0:
            self.primary_action.show()
            self.info_label.show()
            self.remove_all_button.show()

            # Update the file count in the info label
            total_size_bytes = 0
            for index in range(self.file_selection.file_list.count()):
                item = self.file_selection.file_list.item(index)
                total_size_bytes += item.size_bytes
            total_size_readable = self.common.human_readable_filesize(total_size_bytes)

            if file_count > 1:
                self.info_label.setText(
                    strings._("gui_file_info").format(file_count, total_size_readable)
                )
            else:
                self.info_label.setText(
                    strings._("gui_file_info_single").format(
                        file_count, total_size_readable
                    )
                )

        else:
            self.primary_action.hide()
            self.info_label.hide()
            self.remove_all_button.hide()

    def reset_info_counters(self):
        """
        Set the info counters back to zero.
        """
        self.history.reset()
        self.toggle_history.indicator_count = 0
        self.toggle_history.update_indicator()

    def delete_all(self):
        """
        Delete All button clicked
        """
        self.file_selection.file_list.clear()
        self.file_selection.file_list.files_updated.emit()

        self.file_selection.file_list.setCurrentItem(None)

    @staticmethod
    def _compute_total_size(filenames):
        total_size = 0
        for filename in filenames:
            if os.path.isfile(filename):
                total_size += os.path.getsize(filename)
            if os.path.isdir(filename):
                total_size += Common.dir_size(filename)
        return total_size



# -*- coding: utf-8 -*-
"""
OnionShare | https://onionshare.org/

Copyright (C) 2014-2022 Micah Lee, et al. <micah@micahflee.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from .censorship import CensorshipCircumvention
from .meek import Meek
from stem.control import Controller
from stem import ProtocolError, SocketClosed
from stem.connection import MissingPassword, UnreadableCookieFile, AuthenticationFailure
import base64
import nacl.public
import os
import psutil
import re
import shlex
import subprocess
import tempfile
import time
import traceback

from packaging.version import Version


class TorErrorAutomatic(Exception):
    """
    OnionShare is failing to connect and authenticate to the Tor controller,
    using automatic settings that should work with Tor Browser.
    """


class TorErrorInvalidSetting(Exception):
    """
    This exception is raised if the settings just don't make sense.
    """


class TorErrorSocketPort(Exception):
    """
    OnionShare can't connect to the Tor controller using the supplied address and port.
    """


class TorErrorSocketFile(Exception):
    """
    OnionShare can't connect to the Tor controller using the supplied socket file.
    """


class TorErrorMissingPassword(Exception):
    """
    OnionShare connected to the Tor controller, but it requires a password.
    """


class TorErrorUnreadableCookieFile(Exception):
    """
    OnionShare connected to the Tor controller, but your user does not have permission
    to access the cookie file.
    """


class TorErrorAuthError(Exception):
    """
    OnionShare connected to the address and port, but can't authenticate. It's possible
    that a Tor controller isn't listening on this port.
    """


class TorErrorProtocolError(Exception):
    """
    This exception is raised if onionshare connects to the Tor controller, but it
    isn't acting like a Tor controller (such as in Whonix).
    """


class TorTooOldEphemeral(Exception):
    """
    This exception is raised if the version of tor doesn't support ephemeral onion services
    """


class TorTooOldStealth(Exception):
    """
    This exception is raised if the version of tor doesn't support stealth onion services
    """


class BundledTorTimeout(Exception):
    """
    This exception is raised if onionshare is set to use the bundled Tor binary,
    but Tor doesn't finish connecting promptly.
    """


class BundledTorCanceled(Exception):
    """
    This exception is raised if onionshare is set to use the bundled Tor binary,
    and the user cancels connecting to Tor
    """


class BundledTorBroken(Exception):
    """
    This exception is raised if onionshare is set to use the bundled Tor binary,
    but the process seems to fail to run.
    """


class PortNotAvailable(Exception):
    """
    There are no available ports for OnionShare to use, which really shouldn't ever happen
    """


class Onion(object):
    """
    Onion is an abstraction layer for connecting to the Tor control port and
    creating onion services. OnionShare supports creating onion services by
    connecting to the Tor controller and using ADD_ONION, DEL_ONION.

    stealth: Should the onion service be stealth?

    settings: A Settings object. If it's not passed in, load from disk.

    bundled_connection_func: If the tor connection type is bundled, optionally
    call this function and pass in a status string while connecting to tor. This
    is necessary for status updates to reach the GUI.
    """

    def __init__(self, common, use_tmp_dir=False, get_tor_paths=None):
        self.common = common
        self.common.log("Onion", "__init__")

        self.use_tmp_dir = use_tmp_dir

        # Set the path of the tor binary, for bundled tor
        if not get_tor_paths:
            get_tor_paths = self.common.get_tor_paths
        (
            self.tor_path,
            self.tor_geo_ip_file_path,
            self.tor_geo_ipv6_file_path,
            self.obfs4proxy_file_path,
            self.snowflake_file_path,
            self.meek_client_file_path,
        ) = get_tor_paths()

        # The tor process
        self.tor_proc = None

        # The Tor controller
        self.c = None

        # Start out not connected to Tor
        self.connected_to_tor = False

        # Assigned later if we are using stealth mode
        self.auth_string = None

        # Keep track of onions where it's important to gracefully close to prevent truncated downloads
        self.graceful_close_onions = []

    def key_str(self, key):
        """
        Returns a base32 decoded string of a key.
        """
        # bytes to base 32
        key_bytes = bytes(key)
        key_b32 = base64.b32encode(key_bytes)
        # strip trailing ====
        assert key_b32[-4:] == b"===="
        key_b32 = key_b32[:-4]
        # change from b'ASDF' to ASDF
        s = key_b32.decode("utf-8")
        return s

    def connect(
        self,
        custom_settings=None,
        config=None,
        tor_status_update_func=None,
        connect_timeout=120,
        local_only=False,
    ):
        if local_only:
            self.common.log(
                "Onion", "connect", "--local-only, so skip trying to connect"
            )
            return

        # Either use settings that are passed in, or use them from common
        if custom_settings:
            self.settings = custom_settings
        elif config:
            self.common.load_settings(config)
            self.settings = self.common.settings
        else:
            self.common.load_settings()
            self.settings = self.common.settings

        self.common.log(
            "Onion",
            "connect",
            f"connection_type={self.settings.get('connection_type')}",
        )

        # The Tor controller
        self.c = None

        if self.settings.get("connection_type") == "bundled":
            # Create a torrc for this session
            if self.use_tmp_dir:
                self.tor_data_directory = tempfile.TemporaryDirectory(
                    dir=self.common.build_tmp_dir()
                )
                self.tor_data_directory_name = self.tor_data_directory.name
            else:
                self.tor_data_directory_name = self.common.build_tor_dir()
            self.common.log(
                "Onion",
                "connect",
                f"tor_data_directory_name={self.tor_data_directory_name}",
            )

            # Create the torrc
            with open(self.common.get_resource_path("torrc_template")) as f:
                torrc_template = f.read()
            self.tor_cookie_auth_file = os.path.join(
                self.tor_data_directory_name, "cookie"
            )
            try:
                self.tor_socks_port = self.common.get_available_port(1000, 65535)
            except Exception:
                print("OnionShare port not available")
                raise PortNotAvailable()
            self.tor_torrc = os.path.join(self.tor_data_directory_name, "torrc")

            # If there is an existing OnionShare tor process, kill it
            for proc in psutil.process_iter(["pid", "name", "username"]):
                try:
                    cmdline = proc.cmdline()
                    if (
                        cmdline[0] == self.tor_path
                        and cmdline[1] == "-f"
                        and cmdline[2] == self.tor_torrc
                    ):
                        self.common.log(
                            "Onion", "connect", "found a stale tor process, killing it"
                        )
                        proc.terminate()
                        proc.wait()
                        break
                except Exception:
                    pass

            if self.common.platform == "Windows" or self.common.platform == "Darwin":
                # Windows doesn't support unix sockets, so it must use a network port.
                # macOS can't use unix sockets either because socket filenames are limited to
                # 100 chars, and the macOS sandbox forces us to put the socket file in a place
                # with a really long path.
                torrc_template += "ControlPort {{control_port}}\n"
                try:
                    self.tor_control_port = self.common.get_available_port(1000, 65535)
                except Exception:
                    print("OnionShare port not available")
                    raise PortNotAvailable()
                self.tor_control_socket = None
            else:
                # Linux and BSD can use unix sockets
                torrc_template += "ControlSocket {{control_socket}}\n"
                self.tor_control_port = None
                self.tor_control_socket = os.path.join(
                    self.tor_data_directory_name, "control_socket"
                )

            torrc_template = torrc_template.replace(
                "{{data_directory}}", self.tor_data_directory_name
            )
            torrc_template = torrc_template.replace(
                "{{control_port}}", str(self.tor_control_port)
            )
            torrc_template = torrc_template.replace(
                "{{control_socket}}", str(self.tor_control_socket)
            )
            torrc_template = torrc_template.replace(
                "{{cookie_auth_file}}", self.tor_cookie_auth_file
            )
            torrc_template = torrc_template.replace(
                "{{geo_ip_file}}", self.tor_geo_ip_file_path
            )
            torrc_template = torrc_template.replace(
                "{{geo_ipv6_file}}", self.tor_geo_ipv6_file_path
            )
            torrc_template = torrc_template.replace(
                "{{socks_port}}", str(self.tor_socks_port)
            )
            torrc_template = torrc_template.replace(
                "{{obfs4proxy_path}}", str(self.obfs4proxy_file_path)
            )
            torrc_template = torrc_template.replace(
                "{{snowflake_path}}", str(self.snowflake_file_path)
            )

            with open(self.tor_torrc, "w") as f:
                self.common.log("Onion", "connect", "Writing torrc template file")
                f.write(torrc_template)

                # Bridge support
                if self.settings.get("bridges_enabled"):
                    f.write("\nUseBridges 1\n")
                    if self.settings.get("bridges_type") == "built-in":
                        use_torrc_bridge_templates = False
                        builtin_bridge_type = self.settings.get("bridges_builtin_pt")
                        # Use built-inbridges stored in settings, if they are there already.
                        # They are probably newer than that of our hardcoded copies.
                        if self.settings.get("bridges_builtin"):
                            try:
                                for line in self.settings.get("bridges_builtin")[
                                    builtin_bridge_type
                                ]:
                                    if line.strip() != "":
                                        f.write(f"Bridge {line}\n")
                                self.common.log(
                                    "Onion",
                                    "connect",
                                    "Wrote in the built-in bridges from OnionShare settings",
                                )
                            except KeyError:
                                # Somehow we had built-in bridges in our settings, but
                                # not for this bridge type. Fall back to using the hard-
                                # coded templates.
                                use_torrc_bridge_templates = True
                        else:
                            use_torrc_bridge_templates = True
                        if use_torrc_bridge_templates:
                            if builtin_bridge_type == "obfs4":
                                with open(
                                    self.common.get_resource_path(
                                        "torrc_template-obfs4"
                                    )
                                ) as o:
                                    f.write(o.read())
                            elif builtin_bridge_type == "meek-azure":
                                with open(
                                    self.common.get_resource_path(
                                        "torrc_template-meek_lite_azure"
                                    )
                                ) as o:
                                    f.write(o.read())
                            elif builtin_bridge_type == "snowflake":
                                with open(
                                    self.common.get_resource_path(
                                        "torrc_template-snowflake"
                                    )
                                ) as o:
                                    f.write(o.read())
                            self.common.log(
                                "Onion",
                                "connect",
                                "Wrote in the built-in bridges from torrc templates",
                            )
                    elif self.settings.get("bridges_type") == "moat":
                        for line in self.settings.get("bridges_moat").split("\n"):
                            if line.strip() != "":
                                f.write(f"Bridge {line}\n")

                    elif self.settings.get("bridges_type") == "custom":
                        for line in self.settings.get("bridges_custom").split("\n"):
                            if line.strip() != "":
                                f.write(f"Bridge {line}\n")

            # Execute a tor subprocess
            self.common.log("Onion", "connect", f"starting {self.tor_path} subprocess")
            start_ts = time.time()
            if self.common.platform == "Windows":
                # In Windows, hide console window when opening tor.exe subprocess
                startupinfo = subprocess.STARTUPINFO()
                startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
                self.tor_proc = subprocess.Popen(
                    [self.tor_path, "-f", self.tor_torrc],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    startupinfo=startupinfo,
                )
            else:
                if self.common.is_snapcraft():
                    env = None
                else:
                    env = {"LD_LIBRARY_PATH": os.path.dirname(self.tor_path)}

                self.tor_proc = subprocess.Popen(
                    [self.tor_path, "-f", self.tor_torrc],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    env=env,
                )

            # Wait for the tor controller to start
            self.common.log("Onion", "connect", f"tor pid: {self.tor_proc.pid}")
            time.sleep(2)

            return_code = self.tor_proc.poll()
            if return_code != None:
                self.common.log("Onion", "connect", f"tor process has terminated early: {return_code}")

            # Connect to the controller
            self.common.log("Onion", "connect", "authenticating to tor controller")
            try:
                if (
                    self.common.platform == "Windows"
                    or self.common.platform == "Darwin"
                ):
                    self.c = Controller.from_port(port=self.tor_control_port)
                    self.c.authenticate()
                else:
                    self.c = Controller.from_socket_file(path=self.tor_control_socket)
                    self.c.authenticate()
            except Exception as e:
                print("OnionShare could not connect to Tor:\n{}".format(e.args[0]))
                print(traceback.format_exc())
                raise BundledTorBroken(e.args[0])

            while True:
                try:
                    res = self.c.get_info("status/bootstrap-phase")
                except SocketClosed:
                    raise BundledTorCanceled()

                res_parts = shlex.split(res)
                progress = res_parts[2].split("=")[1]
                summary = res_parts[4].split("=")[1]

                # "\033[K" clears the rest of the line
                print(
                    f"\rConnecting to the Tor network: {progress}% - {summary}\033[K",
                    end="",
                )

                if callable(tor_status_update_func):
                    if not tor_status_update_func(progress, summary):
                        # If the dialog was canceled, stop connecting to Tor
                        self.common.log(
                            "Onion",
                            "connect",
                            "tor_status_update_func returned false, canceling connecting to Tor",
                        )
                        print()
                        return False

                if summary == "Done":
                    print("")
                    break
                time.sleep(0.2)

                # If using bridges, it might take a bit longer to connect to Tor
                if self.settings.get("bridges_enabled"):
                    # Only override timeout if a custom timeout has not been passed in
                    if connect_timeout == 120:
                        connect_timeout = 150
                if time.time() - start_ts > connect_timeout:
                    print("")
                    try:
                        self.tor_proc.terminate()
                        print(
                            "Taking too long to connect to Tor. Maybe you aren't connected to the Internet, or have an inaccurate system clock?"
                        )
                        raise BundledTorTimeout()
                    except FileNotFoundError:
                        pass

        elif self.settings.get("connection_type") == "automatic":
            # Automatically try to guess the right way to connect to Tor Browser
            automatic_error = "Could not connect to the Tor controller. Is Tor Browser (available from torproject.org) running in the background?"

            # Try connecting to control port
            found_tor = False

            # If the TOR_CONTROL_PORT environment variable is set, use that
            env_port = os.environ.get("TOR_CONTROL_PORT")
            if env_port:
                try:
                    self.c = Controller.from_port(port=int(env_port))
                    found_tor = True
                except Exception:
                    pass

            else:
                # Otherwise, try default ports for Tor Browser, Tor Messenger, and system tor
                try:
                    ports = [9151, 9153, 9051]
                    for port in ports:
                        self.c = Controller.from_port(port=port)
                        found_tor = True
                except Exception:
                    pass

                # If this still didn't work, try guessing the default socket file path
                socket_file_path = ""
                if not found_tor:
                    try:
                        if self.common.platform == "Darwin":
                            socket_file_path = os.path.expanduser(
                                "~/Library/Application Support/TorBrowser-Data/Tor/control.socket"
                            )

                        self.c = Controller.from_socket_file(path=socket_file_path)
                        found_tor = True
                    except Exception:
                        pass

            # If connecting to default control ports failed, so let's try
            # guessing the socket file name next
            if not found_tor:
                try:
                    if self.common.platform == "Linux" or self.common.platform == "BSD":
                        socket_file_path = (
                            f"/run/user/{os.geteuid()}/Tor/control.socket"
                        )
                    elif self.common.platform == "Darwin":
                        socket_file_path = (
                            f"/run/user/{os.geteuid()}/Tor/control.socket"
                        )
                    elif self.common.platform == "Windows":
                        # Windows doesn't support unix sockets
                        print(automatic_error)
                        raise TorErrorAutomatic()

                    self.c = Controller.from_socket_file(path=socket_file_path)

                except Exception:
                    print(automatic_error)
                    raise TorErrorAutomatic()

            # Try authenticating
            try:
                self.c.authenticate()
            except Exception:
                print(automatic_error)
                raise TorErrorAutomatic()

        else:
            # Use specific settings to connect to tor
            invalid_settings_error = "Can't connect to Tor controller because your settings don't make sense."

            # Try connecting
            try:
                if self.settings.get("connection_type") == "control_port":
                    self.c = Controller.from_port(
                        address=self.settings.get("control_port_address"),
                        port=self.settings.get("control_port_port"),
                    )
                elif self.settings.get("connection_type") == "socket_file":
                    self.c = Controller.from_socket_file(
                        path=self.settings.get("socket_file_path")
                    )
                else:
                    print(invalid_settings_error)
                    raise TorErrorInvalidSetting()

            except Exception:
                if self.settings.get("connection_type") == "control_port":
                    print(
                        "Can't connect to the Tor controller at {}:{}.".format(
                            self.settings.get("control_port_address"),
                            self.settings.get("control_port_port"),
                        )
                    )
                    raise TorErrorSocketPort(
                        self.settings.get("control_port_address"),
                        self.settings.get("control_port_port"),
                    )
                print(
                    "Can't connect to the Tor controller using socket file {}.".format(
                        self.settings.get("socket_file_path")
                    )
                )
                raise TorErrorSocketFile(self.settings.get("socket_file_path"))

            # Try authenticating
            try:
                if self.settings.get("auth_type") == "no_auth":
                    self.c.authenticate()
                elif self.settings.get("auth_type") == "password":
                    self.c.authenticate(self.settings.get("auth_password"))
                else:
                    print(invalid_settings_error)
                    raise TorErrorInvalidSetting()

            except MissingPassword:
                print(
                    "Connected to Tor controller, but it requires a password to authenticate."
                )
                raise TorErrorMissingPassword()
            except UnreadableCookieFile:
                print(
                    "Connected to the Tor controller, but password may be wrong, or your user is not permitted to read the cookie file."
                )
                raise TorErrorUnreadableCookieFile()
            except AuthenticationFailure:
                print(
                    "Connected to {}:{}, but can't authenticate. Maybe this isn't a Tor controller?".format(
                        self.settings.get("control_port_address"),
                        self.settings.get("control_port_port"),
                    )
                )
                raise TorErrorAuthError(
                    self.settings.get("control_port_address"),
                    self.settings.get("control_port_port"),
                )

        # If we made it this far, we should be connected to Tor
        self.connected_to_tor = True

        # Get the tor version
        self.tor_version = self.c.get_version().version_str
        self.common.log("Onion", "connect", f"Connected to tor {self.tor_version}")

        # Do the versions of stem and tor that I'm using support ephemeral onion services?
        list_ephemeral_hidden_services = getattr(
            self.c, "list_ephemeral_hidden_services", None
        )
        self.supports_ephemeral = (
            callable(list_ephemeral_hidden_services) and self.tor_version >= "0.2.7.1"
        )

        # Do the versions of stem and tor that I'm using support v3 stealth onion services?
        try:
            res = self.c.create_ephemeral_hidden_service(
                {1: 1},
                basic_auth=None,
                await_publication=False,
                key_type="NEW",
                key_content="ED25519-V3",
                client_auth_v3="E2GOT5LTUTP3OAMRCRXO4GSH6VKJEUOXZQUC336SRKAHTTT5OVSA",
            )
            tmp_service_id = res.service_id
            self.c.remove_ephemeral_hidden_service(tmp_service_id)
            self.supports_stealth = True
        except Exception:
            # ephemeral stealth onion services are not supported
            self.supports_stealth = False

        # Does this version of Tor support next-gen ('v3') onions?
        # Note, this is the version of Tor where this bug was fixed:
        # https://trac.torproject.org/projects/tor/ticket/28619
        cleaned_tor_version = re.sub(r"\s*\(.*\)", "", self.tor_version)
        self.supports_v3_onions = Version(cleaned_tor_version) >= Version("0.3.5.7")

        # Now that we are connected to Tor, if we are using built-in bridges,
        # update them with the latest copy available from the Tor API
        if (
            self.settings.get("bridges_enabled")
            and self.settings.get("bridges_type") == "built-in"
        ):
            self.update_builtin_bridges()

    def is_authenticated(self):
        """
        Returns True if the Tor connection is still working, or False otherwise.
        """
        if self.c is not None:
            return self.c.is_authenticated()
        else:
            return False

    def start_onion_service(self, mode, mode_settings, port, await_publication):
        """
        Start a onion service on port 80, pointing to the given port, and
        return the onion hostname.
        """
        self.common.log("Onion", "start_onion_service", f"port={port}")

        if not self.supports_ephemeral:
            print(
                "Your version of Tor is too old, ephemeral onion services are not supported"
            )
            raise TorTooOldEphemeral()

        if mode_settings.get("onion", "private_key"):
            key_content = mode_settings.get("onion", "private_key")
            key_type = "ED25519-V3"
        else:
            key_content = "ED25519-V3"
            key_type = "NEW"

        debug_message = f"key_type={key_type}"
        if key_type == "NEW":
            debug_message += f", key_content={key_content}"
        self.common.log("Onion", "start_onion_service", debug_message)

        if mode_settings.get("general", "public"):
            client_auth_priv_key = None
            client_auth_pub_key = None
        else:
            if not self.supports_stealth:
                print(
                    "Your version of Tor is too old, stealth onion services are not supported"
                )
                raise TorTooOldStealth()
            else:
                if key_type == "NEW" or not mode_settings.get(
                    "onion", "client_auth_priv_key"
                ):
                    # Generate a new key pair for Client Auth on new onions, or if
                    # it's a persistent onion but for some reason we don't them
                    client_auth_priv_key_raw = nacl.public.PrivateKey.generate()
                    client_auth_priv_key = self.key_str(client_auth_priv_key_raw)
                    client_auth_pub_key = self.key_str(
                        client_auth_priv_key_raw.public_key
                    )
                else:
                    # These should have been saved in settings from the previous run of a persistent onion
                    client_auth_priv_key = mode_settings.get(
                        "onion", "client_auth_priv_key"
                    )
                    client_auth_pub_key = mode_settings.get(
                        "onion", "client_auth_pub_key"
                    )

        try:
            if not self.supports_stealth:
                res = self.c.create_ephemeral_hidden_service(
                    {80: port},
                    await_publication=await_publication,
                    basic_auth=None,
                    key_type=key_type,
                    key_content=key_content,
                )
            else:
                res = self.c.create_ephemeral_hidden_service(
                    {80: port},
                    await_publication=await_publication,
                    basic_auth=None,
                    key_type=key_type,
                    key_content=key_content,
                    client_auth_v3=client_auth_pub_key,
                )

        except ProtocolError as e:
            print("Tor error: {}".format(e.args[0]))
            raise TorErrorProtocolError(e.args[0])

        onion_host = res.service_id + ".onion"

        # Gracefully close share mode rendezvous circuits
        if mode == "share":
            self.graceful_close_onions.append(res.service_id)

        # Save the service_id
        mode_settings.set("general", "service_id", res.service_id)

        # Save the private key and hidservauth string
        if not mode_settings.get("onion", "private_key"):
            mode_settings.set("onion", "private_key", res.private_key)

        # If using V3 onions and Client Auth, save both the private and public key
        # because we need to send the public key to ADD_ONION (if we restart this
        # same share at a later date), and the private key to the other user for
        # their Tor Browser.
        if not mode_settings.get("general", "public"):
            mode_settings.set("onion", "client_auth_priv_key", client_auth_priv_key)
            mode_settings.set("onion", "client_auth_pub_key", client_auth_pub_key)
            # If we were pasting the client auth directly into the filesystem behind a Tor client,
            # it would need to be in the format below. However, let's just set the private key
            # by itself, as this can be pasted directly into Tor Browser, which is likely to
            # be the most common use case.
            # self.auth_string = f"{onion_host}:x25519:{client_auth_priv_key}"
            self.auth_string = client_auth_priv_key

        return onion_host

    def stop_onion_service(self, mode_settings):
        """
        Stop a specific onion service
        """
        onion_host = mode_settings.get("general", "service_id")
        if onion_host:
            self.common.log("Onion", "stop_onion_service", f"onion host: {onion_host}")
            try:
                self.c.remove_ephemeral_hidden_service(
                    mode_settings.get("general", "service_id")
                )
            except Exception:
                self.common.log(
                    "Onion", "stop_onion_service", f"failed to remove {onion_host}"
                )

    def cleanup(self, stop_tor=True, wait=True):
        """
        Stop onion services that were created earlier. If there's a tor subprocess running, kill it.
        """
        self.common.log("Onion", "cleanup")

        # Cleanup the ephemeral onion services, if we have any
        try:
            onions = self.c.list_ephemeral_hidden_services()
            for service_id in onions:
                onion_host = f"{service_id}.onion"
                try:
                    self.common.log(
                        "Onion", "cleanup", f"trying to remove onion {onion_host}"
                    )
                    self.c.remove_ephemeral_hidden_service(service_id)
                except Exception:
                    self.common.log(
                        "Onion", "cleanup", f"failed to remove onion {onion_host}"
                    )
                    pass
        except Exception:
            pass

        if stop_tor:
            # Stop tor process
            if self.tor_proc:
                if wait:
                    # Wait for Tor rendezvous circuits to close
                    # Catch exceptions to prevent crash on Ctrl-C
                    try:
                        rendezvous_circuit_ids = []
                        for c in self.c.get_circuits():
                            if (
                                c.purpose == "HS_SERVICE_REND"
                                and c.rend_query in self.graceful_close_onions
                            ):
                                rendezvous_circuit_ids.append(c.id)

                        symbols = list("\\|/-")
                        symbols_i = 0

                        while True:
                            num_rend_circuits = 0
                            for c in self.c.get_circuits():
                                if c.id in rendezvous_circuit_ids:
                                    num_rend_circuits += 1

                            if num_rend_circuits == 0:
                                print(
                                    "\rTor rendezvous circuits have closed" + " " * 20
                                )
                                break

                            if num_rend_circuits == 1:
                                circuits = "circuit"
                            else:
                                circuits = "circuits"
                            print(
                                f"\rWaiting for {num_rend_circuits} Tor rendezvous {circuits} to close {symbols[symbols_i]} ",
                                end="",
                            )
                            symbols_i = (symbols_i + 1) % len(symbols)
                            time.sleep(1)
                    except Exception:
                        pass

                self.tor_proc.terminate()
                time.sleep(0.2)
                if self.tor_proc.poll() is None:
                    self.common.log(
                        "Onion",
                        "cleanup",
                        "Tried to terminate tor process but it's still running",
                    )
                    try:
                        self.tor_proc.kill()
                        time.sleep(0.2)
                        if self.tor_proc.poll() is None:
                            self.common.log(
                                "Onion",
                                "cleanup",
                                "Tried to kill tor process but it's still running",
                            )
                    except Exception:
                        self.common.log(
                            "Onion", "cleanup", "Exception while killing tor process"
                        )
                self.tor_proc = None

            # Reset other Onion settings
            self.connected_to_tor = False

            try:
                # Delete the temporary tor data directory
                if self.use_tmp_dir:
                    self.tor_data_directory.cleanup()
            except Exception:
                pass

    def get_tor_socks_port(self):
        """
        Returns a (address, port) tuple for the Tor SOCKS port
        """
        self.common.log("Onion", "get_tor_socks_port")

        if self.settings.get("connection_type") == "bundled":
            return ("127.0.0.1", self.tor_socks_port)
        elif self.settings.get("connection_type") == "automatic":
            return ("127.0.0.1", 9150)
        else:
            return (self.settings.get("socks_address"), self.settings.get("socks_port"))

    def update_builtin_bridges(self):
        """
        Use the CensorshipCircumvention API to fetch the latest built-in bridges
        and update them in settings.
        """
        builtin_bridges = False
        meek = None
        # Try obtaining bridges over Tor, if we're connected to it.
        if self.is_authenticated:
            self.common.log(
                "Onion",
                "update_builtin_bridges",
                "Updating the built-in bridges. Trying over Tor first",
            )
            self.censorship_circumvention = CensorshipCircumvention(
                self.common, None, self
            )
            builtin_bridges = self.censorship_circumvention.request_builtin_bridges()

        if not builtin_bridges:
            # Tor was not running or it failed to hit the Tor API.
            # Fall back to using Meek (domain-fronting).
            self.common.log(
                "Onion",
                "update_builtin_bridges",
                "Updating the built-in bridges. Trying via Meek (no Tor)",
            )
            meek = Meek(self.common)
            meek.start()
            self.censorship_circumvention = CensorshipCircumvention(
                self.common, meek, None
            )
            builtin_bridges = self.censorship_circumvention.request_builtin_bridges()
            meek.cleanup()

        if builtin_bridges:
            # If we got to this point, we have bridges
            self.common.log(
                "Onion",
                "update_builtin_bridges",
                f"Obtained bridges: {builtin_bridges}",
            )
            # Save the new settings
            self.settings.set("bridges_builtin", builtin_bridges)
            self.settings.save()
        else:
            self.common.log(
                "Onion", "update_builtin_bridges", "Error getting built-in bridges"
            )
            return False



# -*- coding: utf-8 -*-
"""
OnionShare | https://onionshare.org/

Copyright (C) 2014-2022 Micah Lee, et al. <micah@micahflee.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import os
import json
import platform

if platform.system() == "Darwin":
    import pwd


class ModeSettings:
    """
    This stores the settings for a single instance of an OnionShare mode. In CLI there
    is only one ModeSettings, and in the GUI there is a separate ModeSettings for each tab
    """

    def __init__(self, common, filename=None, id=None):
        self.common = common

        self.default_settings = {
            "onion": {
                "private_key": None,
                "client_auth_priv_key": None,
                "client_auth_pub_key": None,
            },
            "persistent": {
                "mode": None,
                "enabled": False,
                "autostart_on_launch": False
            },
            "general": {
                "title": None,
                "public": False,
                "autostart_timer": False,
                "autostop_timer": False,
                "service_id": None,
            },
            "share": {
                "autostop_sharing": True,
                "filenames": [],
                "log_filenames": False,
            },
            "receive": {
                "data_dir": self.build_default_receive_data_dir(),
                "webhook_url": None,
                "disable_text": False,
                "disable_files": False,
            },
            "website": {
                "disable_csp": False,
                "custom_csp": None,
                "log_filenames": False,
                "filenames": []
            },
            "chat": {},
        }
        self._settings = {}

        self.just_created = False
        if id:
            self.id = id
        else:
            self.id = self.common.build_password(3)

        self.load(filename)

    def fill_in_defaults(self):
        """
        If there are any missing settings from self._settings, replace them with
        their default values.
        """
        for key in self.default_settings:
            if key in self._settings:
                for inner_key in self.default_settings[key]:
                    if inner_key not in self._settings[key]:
                        self._settings[key][inner_key] = self.default_settings[key][
                            inner_key
                        ]
            else:
                self._settings[key] = self.default_settings[key]

    def get(self, group, key):
        return self._settings[group][key]

    def set(self, group, key, val):
        self._settings[group][key] = val
        self.common.log(
            "ModeSettings", "set", f"updating {self.id}: {group}.{key} = {val}"
        )
        self.save()

    def build_default_receive_data_dir(self):
        """
        Returns the path of the default Downloads directory for receive mode.
        """

        if self.common.platform == "Darwin":
            # We can't use os.path.expanduser() in macOS because in the sandbox it
            # returns the path to the sandboxed homedir
            real_homedir = pwd.getpwuid(os.getuid()).pw_dir
            return os.path.join(real_homedir, "OnionShare")
        elif self.common.platform == "Windows":
            # On Windows, os.path.expanduser() needs to use backslash, or else it
            # retains the forward slash, which breaks opening the folder in explorer.
            return os.path.expanduser("~\\OnionShare")
        else:
            # All other OSes
            return os.path.expanduser("~/OnionShare")

    def load(self, filename=None):
        # Load persistent settings from disk. If the file doesn't exist, create it
        if filename:
            self.filename = filename
        else:
            self.filename = os.path.join(
                self.common.build_persistent_dir(), f"{self.id}.json"
            )

        if os.path.exists(self.filename):
            try:
                with open(self.filename, "r") as f:
                    self._settings = json.load(f)
                    self.fill_in_defaults()
                    self.common.log("ModeSettings", "load", f"loaded {self.filename}")
                    return
            except Exception:
                pass

        # If loading settings didn't work, create the settings file
        self.common.log("ModeSettings", "load", f"creating {self.filename}")
        self.fill_in_defaults()
        self.just_created = True

    def save(self):
        # Save persistent setting to disk
        if not self.get("persistent", "enabled"):
            return

        if self.filename:
            with open(self.filename, "w") as file:
                file.write(json.dumps(self._settings, indent=2))

    def delete(self):
        # Delete the file from disk
        if os.path.exists(self.filename):
            os.remove(self.filename)



# -*- coding: utf-8 -*-
"""
OnionShare | https://onionshare.org/

Copyright (C) 2014-2022 Micah Lee, et al. <micah@micahflee.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import json
import os

strings = {}
translations = {}


def load_strings(common, locale_dir):
    """
    Loads translated strings and fallback to English
    if the translation does not exist.
    """
    global strings, translations

    # Load all translations
    translations = {}
    for locale in common.settings.available_locales:
        filename = os.path.join(locale_dir, f"{locale}.json")
        with open(filename, encoding="utf-8") as f:
            translations[locale] = json.load(f)

    # Build strings
    default_locale = "en"
    current_locale = common.settings.get("locale")
    strings = {}
    for s in translations[default_locale]:
        if (
            current_locale in translations
            and s in translations[current_locale]
            and translations[current_locale][s] != ""
        ):
            strings[s] = translations[current_locale][s]
        else:
            strings[s] = translations[default_locale][s]


def translated(k):
    """
    Returns a translated string.
    """
    return strings[k]


_ = translated



# -*- coding: utf-8 -*-
"""
OnionShare | https://onionshare.org/

Copyright (C) 2014-2022 Micah Lee, et al. <micah@micahflee.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from PySide6 import QtCore


class CompressThread(QtCore.QThread):
    """
    Compresses files to be shared
    """

    success = QtCore.Signal()
    error = QtCore.Signal(str)

    def __init__(self, mode):
        super(CompressThread, self).__init__()
        self.mode = mode
        self.mode.common.log("CompressThread", "__init__")

    # prepare files to share
    def set_processed_size(self, x):
        if self.mode._zip_progress_bar is not None:
            self.mode._zip_progress_bar.update_processed_size_signal.emit(x)

    def run(self):
        self.mode.common.log("CompressThread", "run")

        try:
            self.mode.web.share_mode.set_file_info(
                self.mode.filenames, processed_size_callback=self.set_processed_size
            )
            self.success.emit()
        except OSError as e:
            self.error.emit(e.strerror)

    def cancel(self):
        self.mode.common.log("CompressThread", "cancel")

        # Let the Web and ZipWriter objects know that we're canceling compression early
        self.mode.web.cancel_compression = True
        try:
            self.mode.web.zip_writer.cancel_compression = True
        except AttributeError:
            # we never made it as far as creating a ZipWriter object
            pass
